OBJ
{{
******************************************************************
* SNTP Simple Network Time Protocol                       v1.3   *
*                                                                *
* Copyright (c) 2011 Parallax                                    *
* See end of file for terms of use.                              *
******************************************************************


Revision History:
04-07-2011          - File created

09-08-2011          - Minor code update to correct days in Month rendering
                    - and replace bytefill with bytemove for the 'ref-id' string                               

By R. Forbes
12-07-2012          -Rewrote HumanTime method to address leap year issue. Also, fix incorrect byte addressing in several methods.  
03-31-2013          -Implemented daylightsavings feature.
12-05-2013          -Fixed up comments, straightened out a couple small things. NO KNOWN ISSUES. 
12-29-2013          -Did a final clean up and tested extensively. It works.
}}
Var
long seconds , minutes , hours , day , month , year , dow , daycnt

PUB CreateUDPtimeheader(BufferAddress,IPAddr)
  '---------------------------------------------------------------------
  '                     UDP IP Address - 4 Bytes 
  '---------------------------------------------------------------------
    BYTEMOVE(BufferAddress,IPAddr,4)
             'Destination value count
  '---------------------------------------------------------------------
  '                       UDP Header - 4 Bytes 
  '---------------------------------------------------------------------
    byte[BufferAddress][4] := 0
    byte[BufferAddress][5] := 123 '<- Port Address 
    byte[BufferAddress][6] := 0 
    byte[BufferAddress][7] := 48  '<- Header + Packet
  '---------------------------------------------------------------------
  '                       UDP Packet - 44 Bytes
  '---------------------------------------------------------------------
    byte[BufferAddress][8]  := %11_100_011    'leap,version, and mode
    byte[BufferAddress][9]  := 0              'stratum
    byte[BufferAddress][10] := 0              'Poll   
    byte[BufferAddress][11] := %10010100      'precision
    byte[BufferAddress][12] := 0              'rootdelay
    byte[BufferAddress][13] := 0              'rootdelay   
    byte[BufferAddress][14] := 0              'rootdispersion
    byte[BufferAddress][15] := 0              'rootdispersion

    bytemove(BufferAddress+16,string("LOCL"),4) 'ref-id ; four-character ASCII string

    bytefill(BufferAddress+20,0,32)           '(ref, originate, receive, transmit) time 

{

byte 8=        leap           = %11           ; alarm condition (clock not synchronized) 
byte 8=        version        = %011 or %100  ; Version 3 or 4
byte 8=        mode           = %011          ; Client 
byte 9=        stratum        = %00000000     ; unspecified 
byte 10=       Poll           = %00000000     ; = 2^n seconds (maximum interval between successive messages)
byte 11=       precision      = %10010100     ; -20 (8-bit signed integer)
byte 12 to 15= rootdelay      = 0             ; 32 bit value
byte 16 to 19= rootdispersion = 0             ; 32 bit value
byte 20 to 23= ref id         = "LOCL"        ; four-character ASCII string
byte 24 to 31= ref time       = 0             ; 64 bit value (8 bytes, 2 longs each)
byte 32 to 39= originate time = 0             ; 64 bit value   
byte 40 to 47= receive time   = 0             ; 64 bit value
byte 48 to 56= transmit time  = 0             ; 64 bit value
}

PUB GetLI(BufferAddress)
    result := (byte[BufferAddress][8] & %11000000)>>6

    ' first two bits of byte 0
    '0 - No warning
    '1 - last minute has 61 seconds
    '2 - last minute has 59 seconds
    '3 - alarm condition (clock not synchronized)   

PUB GetVersion(BufferAddress)    
    result := (byte[BufferAddress][8] & %00111000)>>3

    'bits 2,3,4 of byte 0
    '3 - Version 3 (IPv4 only)
    '4 - Version 4 (IPv4, IPv6 and OSI)

PUB GetMode(BufferAddress)

    result := byte[BufferAddress][8] & %00000111      

    'bits 0,1 and 2 of byte 0
    '0 - reserved
    '1 - symmetric active
    '2 - symmetric passive
    '3 - client
    '4 - server
    '5 - broadcast
    '6 - reserved for NTP control message
    '7 - reserved for private use

PUB GetStratum(BufferAddress)

     result := byte[BufferAddress][9] & %11111111
 
    '0      - unspecified or unavailable
    '1      - primary reference (e.g., radio clock)
    '2-15   - secondary reference (via NTP or SNTP) 
    '16-255 - reserved

PUB GetPoll(BufferAddress)

    result := ~byte[BufferAddress][10]

    'This is an eight-bit signed integer indicating the
    'maximum interval between successive messages, in seconds
    'to the nearest power of two. The values that can appear
    'in this field presently range from 4 (16 s) to 14 (16384 s);
    'however, most applications use only the sub-range 6 (64 s)
    'to 10 (1024 s). 

PUB GetPrecision(BufferAddress)

    result := ~byte[BufferAddress][11]  

    'This is an eight-bit signed integer indicating the
    'precision of the local clock, in seconds to the nearest
    'power of two. The values that normally appear in this
    'field range from -6 for mains-frequency clocks to -20 for
    'microsecond clocks found in some workstations.

PUB GetRootDelay(BufferAddress,frac)|Temp1, Temp2    

    Temp1:=Temp2:=0
    
    Temp1 := byte[BufferAddress][12]<<8 + byte[BufferAddress][13]
    Temp1 := ~~Temp1    
    Temp2 := byte[BufferAddress][14]<<8 + byte[BufferAddress][15]
    Temp2 := ~~Temp2
    if frac
          return Temp2

    return Temp1
    
    'This is a 32-bit signed fixed-point number indicating the
    'total roundtrip delay to the primary reference source, in
    'seconds with fraction point between bits 15 and 16. Note
    'that this variable can take on both positive and negative
    'values, depending on the relative time and frequency offsets.
    'The values that normally appear in this field range from
    'negative values of a few milliseconds to positive values of
    'several hundred milliseconds.

PUB GetRootDispersion(BufferAddress,frac)|Temp1, Temp2  

    Temp1:=Temp2:=0
      
    Temp1 := byte[BufferAddress][16]<<8 + byte[BufferAddress][17]
    Temp1 := ~~Temp1
    Temp2 := byte[BufferAddress][18]<<8 + byte[BufferAddress][19]
    Temp2 := ~~Temp2 
    if frac
          return Temp2

    return Temp1

    'This is a 32-bit unsigned fixed-point number indicating the
    'nominal error relative to the primary reference source, in
    'seconds with fraction point between bits 15 and 16. The values
    'that normally appear in this field range from 0 to several
    'hundred milliseconds. 

PUB GetReferenceIdentifier(BufferAddress,FillAddress)|temp
{
      Calling example:          
            pst.str(GetReferenceIdentifier(@rxdata,string("----")))

            dashes get replaced with 4-Character Buffer contents
         
}   

    bytemove(FillAddress,BufferAddress+20,4)                

    result := FillAddress

{          Reference Identifier return codes
       
     +------+----------------------------------------------------------+
     | ID   | Clock Source                                             |
     +------+----------------------------------------------------------+
     | GOES | Geosynchronous Orbit Environment Satellite               |
     | GPS  | Global Position System                                   |
     | GAL  | Galileo Positioning System                               |
     | PPS  | Generic pulse-per-second                                 |
     | IRIG | Inter-Range Instrumentation Group                        |
     | WWVB | LF Radio WWVB Ft. Collins, CO 60 kHz                     |
     | DCF  | LF Radio DCF77 Mainflingen, DE 77.5 kHz                  |
     | HBG  | LF Radio HBG Prangins, HB 75 kHz                         |
     | MSF  | LF Radio MSF Anthorn, UK 60 kHz                          |
     | JJY  | LF Radio JJY Fukushima, JP 40 kHz, Saga, JP 60 kHz       |
     | LORC | MF Radio LORAN C station, 100 kHz                        |
     | TDF  | MF Radio Allouis, FR 162 kHz                             |
     | CHU  | HF Radio CHU Ottawa, Ontario                             |
     | WWV  | HF Radio WWV Ft. Collins, CO                             |
     | WWVH | HF Radio WWVH Kauai, HI                                  |
     | NIST | NIST telephone modem                                     |
     | ACTS | NIST telephone modem                                     |
     | USNO | USNO telephone modem                                     |
     | PTB  | European telephone modem                                 |
     +------+----------------------------------------------------------+          
}
PUB  GetReferenceTimestamp(zone,DLS,BufferAddress, longHIGH,      LongLOW)|Temp1

     Temp1 := byte[BufferAddress][24]<<24+byte[BufferAddress][25]<<16
     Temp1 += byte[BufferAddress][26]<<8 +byte[BufferAddress][27]
     long[LongHIGH]:=Temp1
     Temp1 := byte[BufferAddress][28]<<24+byte[BufferAddress][29]<<16
     Temp1 += byte[BufferAddress][30]<<8 +byte[BufferAddress][31]
     long[LongLOW]:=Temp1     
     'This is the time at which the local clock was
     'last set or corrected, in 64-bit timestamp format.

     HumanTime(zone,DLS,LongHIGH)

PUB  GetOriginateTimestamp(zone,DLS,BufferAddress, longHIGH,      LongLOW)|Temp1

     Temp1 := byte[BufferAddress][32]<<24+byte[BufferAddress][33]<<16
     Temp1 += byte[BufferAddress][34]<<8 +byte[BufferAddress][35]
     long[LongHIGH]:=Temp1
     Temp1 := byte[BufferAddress][36]<<24+byte[BufferAddress][37]<<16
     Temp1 += byte[BufferAddress][38]<<8 +byte[BufferAddress][39]
     long[LongLOW]:=Temp1     
     'This is the time at which the request departed the
     'client for the server, in 64-bit timestamp format.

     HumanTime(zone,DLS,LongHIGH)

PUB  GetReceiveTimestamp(zone,DLS,BufferAddress, longHIGH,      LongLOW)|Temp1

     Temp1 := byte[BufferAddress][40]<<24+byte[BufferAddress][41]<<16
     Temp1 += byte[BufferAddress][42]<<8 +byte[BufferAddress][43]
     long[LongHIGH]:=Temp1
     Temp1 := byte[BufferAddress][44]<<24+byte[BufferAddress][45]<<16
     Temp1 += byte[BufferAddress][46]<<8 +byte[BufferAddress][47]
     long[LongLOW]:=Temp1     
     'This is the time at which the request arrived at
     'the server, in 64-bit timestamp format.

     HumanTime(zone,DLS,LongHIGH)     


PUB  GetTransmitTimestamp(zone,DLS,BufferAddress,longHIGH,LongLOW)   |Temp1
                        
     Temp1 := byte[BufferAddress][48]<<24+byte[BufferAddress][49]<<16
     Temp1 += byte[BufferAddress][50]<<8 +byte[BufferAddress][51]
     long[LongHIGH]:=Temp1
     Temp1 := byte[BufferAddress][52]<<24+byte[BufferAddress][53]<<16
     Temp1 += byte[BufferAddress][54]<<8 +byte[BufferAddress][55]
     long[LongLOW]:=Temp1     
     'This is the time at which the reply departed the
     'server for the client, in 64-bit timestamp format.

     HumanTime(zone,DLS,longHIGH) 
     
PUB HumanTime(zone,DLS,longHIGH)| Totaldays , DSA
'  zone = time zone.  -5 is eastern
'  DLS = DayLightSavings.  1 = use daylight savings, 0= do NOT use daylight savings.
'  longHIGH  seconds value

'By starting with a smaller number of seconds for this loop stuff, we do the loops less times.
'Less loop time= faster completion time.
'That's why the time in this method is based off an epoch of 2012, jan 01 @00:00:00

     DSA := 0                                'Daylight savings adjustment
  
     If DLS                                  'Adjust for daylight savings, if active and used in your time zone.

          DSA := DST(long[longHIGH])

     seconds := long[longHIGH]+DSA

     seconds := seconds - 3_534_364_800      'Adjust the NTP value for 2012: 
                                             '3_534_364_800 is the ntp value for 2012, Jan 01, 0:00:00 hours.         
                                             '2038, jan 19, 03:14:07 is the latest
                                             'date and time supported by NTP at this time (4356472447)
    
     seconds := seconds + (zone * 3600)      'Adjust for your time zone

     year := 2012                            'This is our starting year (modified epoch)

     month := 0

     day := 0

     daycnt := Totaldays := (seconds / 86400)

     'seconds:=seconds-(Totaldays*86400)

     seconds //= 86400

     repeat 

          if (leapyear(year)) 

               repeat month from 1 to 12

                    repeat day from 1 to byte[@DiMLY][month]  
     
                         Totaldays--                           

                         If Totaldays < 0

                              quit 

                    if Totaldays < 0

                         quit

          if NOT (leapyear(year))

               repeat month from 1 to 12

                    repeat day from 1 to byte[@DiM][month] 
     
                         Totaldays--                          

                         If Totaldays < 0

                              quit 

                    if Totaldays < 0

                         quit

          if Totaldays < 0

           quit

          year++

'==========================================
  'Once the loop above completes, we'll be left with less than a days worth of seconds.
  'So, let's see how many hours, minutes and seconds we have left.
{
  hours:=0

  minutes:=0

  repeat while (seconds=>60)

     seconds-=60

     minutes++

     if minutes == 60

        hours++

        minutes:=0
}

     hours := seconds / 3600

     seconds//= 3600

     minutes := seconds / 60

     seconds//= 60
     
'=========================================
  'What's the day of the week? (Assuming Sunday== 1st day of week)

     dow := ((daycnt) // 7) + 1

   'if daycnt=6     6//7 = 6 + 1 = 7
   '          7     7//7 = 0 + 1 = 1
   '          8     8//7   1 + 1 = 2
   '          9     9//7   2 + 1 = 3
   '          10    10//7  3 + 1 = 4
   '          11    11//7  4 + 1 = 5
   '          12    12//7  5 + 1 = 6
   '          13    13//7  6 + 1 = 7
   '          14    14//7  0 + 1 = 1
                        
'=========================================
  'Once the loop above completes, we have our time and date data ready to use.

     long[longHIGH][2] := month<<24 + day<<16 + year

     long[longHIGH][3] := dow<<24 + hours<<16 + minutes<<8 + seconds 

pri leapyear(check)
{
  Case check

    2012: return 1   'Enter more leap years to extend the available range of this calculator.
    2016: return 1
    2020: return 1
    2024: return 1
    2028: return 1
    2032: return 1
    2036: return 1
    2040: return 1
}

     if NOT check // 4  

          return 1

     return 0     

    
pri DST(second)|adjustment

   
  Case second

    3571869600..3592432800: adjustment := 3600         '2013
    3603319200..3623882400: adjustment := 3600         '2014
    3634941600..3655504800: adjustment := 3600         'etc
    3666823200..3687386400: adjustment := 3600         'etc
    3698272800..3718836000: adjustment := 3600 
    3729722400..3750285600: adjustment := 3600 
    3761172000..3781735200: adjustment := 3600 
    3792621600..3813184800: adjustment := 3600         '2020
    other: adjustment := 0

  return adjustment
  
      
{
                  DST = second Sunday in March to the first Sunday in November
                  For 2013- In March, we move Forward 1 hour on Mar 10 at 2 AM and then
                  back one hour on Nov 3 at 2 AM. 
                                           NTP time values
                  2013- Mar 10 to Nov 3    3571869600 - 3592432800
                  2014- Mar 9  to Nov 2    3603319200 - 3623882400
                  2015- Mar 10 to Nov 3    3634941600 - 3655504800
                  2016      13        6    3666823200 - 3687386400
                  2017      12        5    3698272800 - 3718836000
                  2018      11        4    3729722400 - 3750285600
                  2019      10        3    3761172000 - 3781735200
                  2020      8         1    3792621600 - 3813184800
                 
                  UTC Time = NTP time - 2208988800
}
DAT
DiMLY  byte  0,31 ,29 ,31 ,30 ,31 ,30 ,31 ,31 ,30 ,31 ,30 ,31
DiM    byte  0,31 ,28 ,31 ,30 ,31 ,30 ,31 ,31 ,30 ,31 ,30 ,31           'Days in Month
              'jan feb mar apr may jun jul aug sep oct nov dec
             'First byte in each array is just a placeholder.